# -*- coding: utf-8 -*-

"""
REALSpaceSearchTool.py - functions for searching REAL space using the Feature Trees (FTrees) tool.

Use of this code is subject to the terms of the StarDrop License Agreement and
use, copying or redistribution of this file for use by other persons or
organisations is prohibited

Copyright(C) Optibrium Ltd 2019
stardrop-support@optibrium.com

The FTrees tool was created and is maintained BioSolveIT. Real Space was created by Enamine.
"""

import PythonQt
import app
import socket
import os
import time
import sys
import glob

from subprocess32_327 import subprocess32 as subprocess
from subprocess import PIPE
import webbrowser

import csv
import REALSpaceCSVHelper
import REALSpaceSearchTool_config
import REALSpaceQueryLogger

log = REALSpaceQueryLogger.REALSpaceQueryLogger('REALSpaceSearchTool.log')
socket.setdefaulttimeout(60)

def search_using_FTrees(appui):
    """
        Perform a search of Enamine REAL space for compounds similar to the selected compound.

        One molecule is provided.
        This tool creates a data set of the similar molecules, gives their REAL space name and
        shows the reagents which can be used to create the compound.

        :param appui:    The stardrop appui object
        :return: None
    """
    try:
        # get currently selected molecule
        SMILES_input = appui.getCurrentMolecules()

        # if no molecule is selected, tell user to select one and re-run
        if not SMILES_input:
            appui.show(
                "Please select a molecule before searching for similar compounds")
        elif len(SMILES_input) > 1:
            appui.show(
                "Please select exactly one molecule")

        else:
            # get the compound ID for the query compound. If none, don't set an ID for the query molecule
            try:
                query_ID = appui.getSelectedIDs()[0]
            except:
                query_ID = ""

            SMILES_input = SMILES_input[0]
            print("Running Enamine REAL space query for " + SMILES_input)

            if os.path.isfile("path_to_REAL_space_file.txt"):
                with open("path_to_REAL_space_file.txt", 'r') as file:
                    space_file = file.read()

            else:
                try:
                    directory = os.getcwd()
                    directory_tuple = directory.split("\\")
                    for item in directory_tuple:
                        if item == "AppData":
                            shortened = directory.split("AppData")
                            downloads_dir = shortened[0] + ("Downloads\\")
                            space_file_concat = glob.glob(downloads_dir + "REALspace_*")
                            space_file = space_file_concat[0]
                except:
                    appui.show('Please browse to REAL space file')
                    browser = PythonQt.PyFileDialog()
                    browser.setWindowTitle('Please browse to REAL space file')
                    path_to_space = browser.run()
                    space_file = path_to_space[0]
                    if not path_to_space:
                        cancel = True

            # Ask user for input parameters
            params = PythonQt.FTreesWidget()
            params.setWindowTitle("Choose REAL Space search parameters")

            if params.run():
                similarity = params.getSpinBoxValue()
                if similarity < 0.8:
                    min_sim_threshold = similarity

                else:
                    min_sim_threshold = 0.8
                similarity = round(similarity, 2)

                num_of_results = params.getMaxNumResults()
                num_of_results = int(num_of_results)

                # warn users if the results will take a very long time to return
                if num_of_results > 99999:
                    warning_dlg = PythonQt.Warning()
                    warning_dlg.setMessage("It may take a long time to return so many compounds, are you sure you would like to continue?")

                    if warning_dlg.run():
                        run_command(appui, SMILES_input, space_file, similarity, min_sim_threshold, num_of_results, query_ID)

                else:
                    run_command(appui, SMILES_input, space_file, similarity, min_sim_threshold, num_of_results, query_ID)

    except Exception as e:
        log.log(repr(e))


def run_command(appui, SMILES_input, space_file, similarity, min_sim_threshold, num_of_results, query_ID):
    """
            Perform a search of Enamine REAL space for compounds similar to the selected compound.

            This tool runs the FTrees program via the command line, then sends the output CSV file to the
            CSV helper to package it into a StarDrop data set.

            :param appui:               The stardrop appui object
            :param SMILES_input         The selected compound
            :param space_file           The Enamine REAL space file
            :param similarity           The target similarity to the selected compound
            :param min_sim_threshold    The minimum similarity which will be returned. If the target similarity
                                        is 0.8 or greater, it will be 0.8, else it will be the same as the
                                        target similarity
            :param num_of_results       The maximum number of results to be returned
            :return: None
    """
    try:
        # show progress bar
        progDlg = PythonQt.SimpleProgress()
        progDlg.setWindowTitle('Finding REAL space compounds')
        progDlg.setMessage('Searching...')
        progDlg.show()
        progDlg.setProgress(8)

        # An SD file is created for the input molecule as this is the format required for the FTrees program
        smiles_file = open("REALSpacereferencemol.sdf", 'wb')
        smiles_file.write(appui.getMolFileFromSmiles(SMILES_input))
        smiles_file.close()

        directory = os.getcwd()
        path_to_FTrees = ""

        # get the path to the FTrees executable, this is stored in a file after a successful execution
        if os.path.isfile("path_to_FTrees_executable.txt"):
            with open("path_to_FTrees_executable.txt", 'r') as file:
                path_to_FTrees = file.read()
        else:
            directory_tuple = directory.split("\\")
            for item in directory_tuple:
                if item == "Roaming":
                    shortened = directory.split("Roaming")
                    try:
                        path_to_BioSolveIT_dir = os.path.join(
                            shortened[0] + "Local\\Programs\\BioSolveIT\\")
                        path_to_FTrees_dir = glob.glob(path_to_BioSolveIT_dir + "FTrees*")
                        path_to_FTrees = path_to_FTrees_dir[-1] + "\\FTrees.exe"
                    except Exception as e:
                        print(e)

        cancel = False
        no_lic_file = False
        current_lic_found = False
        message = ""

        if not path_to_FTrees:
            appui.show('Please browse to FTrees application')
            browser = PythonQt.PyFileDialog()
            browser.setWindowTitle("Please browse to FTrees application")
            path_to_FTrees = browser.run()
            path_to_FTrees = path_to_FTrees[0]
            if not path_to_FTrees:
                cancel = True

        try:
            check_license = subprocess.Popen([path_to_FTrees, "--license-info"], stdout=subprocess.PIPE)

            # If license has expired or isn't in the FTrees directory, try to find another version
            # of FTrees which is licensed and use that for the search.
            for item in check_license.stdout:
                message += item
                if "expired" or "Cannot find license file" in item:
                    shortened = directory.split("Roaming")
                    path_to_BioSolveIT_dir = os.path.join(
                        shortened[0] + "Local\\Programs\\BioSolveIT\\")
                    path_to_FTrees_dir = glob.glob(path_to_BioSolveIT_dir + "FTrees*")
                    if path_to_FTrees_dir > 1:
                        for i in range(len(path_to_FTrees_dir)):
                            if not current_lic_found:
                                path_to_FTrees = path_to_FTrees_dir[i] + "\\FTrees.exe"
                                try:
                                    check_license = subprocess.Popen([path_to_FTrees, "--license-info"], stdout=subprocess.PIPE)
                                    for item in check_license.stdout:
                                        if "expired" in item:
                                            current_lic_found = False
                                        elif "Cannot find license file" in item:
                                            current_lic_found = False
                                        elif "Valid key" in item:
                                            current_lic_found = True
                                except:
                                    current_lic_found = False
                        if current_lic_found == False:
                            cancel = True

            if cancel:
                if "expired" in message:
                    appui.show("License has expired" + "\n\n" + message)
                elif "Cannot find license file" in message:
                    appui.show("No license file found" + "\n\n" + message)
                else:
                    appui.show(message)

        except Exception as e:
            print(e)

        if not cancel and not no_lic_file:

            ref_mol_file = os.path.join(directory + "\\REALSpacereferencemol.sdf")
            try:
                results_file = os.path.join(directory + "\\REALResults_1.csv")
            except Exception as e:
                print(e)

            progress = 10.0

            try:
                info = subprocess.STARTUPINFO()
                info.dwFlags = subprocess.STARTF_USESTDHANDLES | subprocess.STARTF_USESHOWWINDOW
                info.wShowWindow = subprocess.SW_HIDE

                num_of_results = str(num_of_results)
                similarity = str(similarity)
                min_sim_threshold = str(min_sim_threshold)

                check_success = subprocess.Popen(
                    [path_to_FTrees, "-i", ref_mol_file, "-f", space_file, "-o", results_file,
                     "--maxNofResults", num_of_results, "--minSimilarityThreshold", min_sim_threshold, "--targetSimilarity", similarity],
                    startupinfo=info, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

                if check_success.poll() is None:
                    time.sleep(10.0)
                    progress += 2.0
                    progDlg.setProgress(progress)
                    progDlg.setMessage(
                        "Although the REAL Space is vast, the search takes only minutes")
                    progDlg.raise_()
                    for i in range(20):
                        time.sleep(1.0)
                        progress += 1.0
                        progDlg.setProgress(progress)
                else:
                    None
                if check_success.poll() is None:
                    progDlg.setMessage(
                        "This searches from over 11 billion theoretical compounds")
                    progDlg.raise_()
                    for i in range(25):
                        time.sleep(1.0)
                        progress += 1.0
                        progDlg.setProgress(progress)
                else:
                    None
                if check_success.poll() is None:
                    progDlg.setProgress(progress)
                    progDlg.setMessage(
                        "The compounds are virtual, yet >80% can be synthesised by Enamine")
                    progDlg.raise_()
                    for i in range(25):
                        time.sleep(1.0)
                        progress += 1.0
                        progDlg.setProgress(progress)
                else:
                    None

            except Exception as e:
                print(e)
                if "The system cannot find the file specified" in e:
                    appui.show('Cannot locate FTrees application')
                else:
                    appui.show("Error encountered")

            progDlg.setProgress(87)

            if check_success.communicate()[1] == '':
                try:
                    dataset = REALSpaceCSVHelper.createDataSet("REALResults_1_1.csv", REALSpaceSearchTool_config)
                except Exception as e:
                    print(e)

                progDlg.setProgress(95)

                if 'https' in query_ID:
                    query_ID = query_ID.split(">")[1]
                    query_ID = query_ID.split("<")[0]

                if dataset:
                    appui.exportDS(query_ID + " REAL Space search", dataset.pack())
                else:
                    appui.showError("Unable to package the results into a StarDrop data set")

                os.remove(os.path.join(directory + "\\REALResults_1_1.csv"))

                # If the run was successful, write the paths to the FTrees executable and REAL space file to
                # text files for future use.
                with open("path_to_FTrees_executable.txt", 'wb') as FTrees_path:
                    FTrees_path.write(path_to_FTrees)

                with open("path_to_REAL_space_file.txt", 'wb') as file:
                    file.write(space_file)

            else:
                output = check_success.communicate()
                try:
                    if "\r\n" in output[1]:
                        formatted = output[1].replace("\r\n", "")
                        if "ERROR:" in formatted:
                            formatted = formatted.replace(" ERROR: ", "", 1)

                    if "Space does not exist" in output[1]:
                        appui.show("Space file was not found\n\nPlease browse to REAL Space file")
                        browser = PythonQt.PyFileDialog()
                        browser.setWindowTitle('Please browse to REAL space file')
                        path_to_space = browser.run()
                        space_file = path_to_space[0]
                        with open("path_to_REAL_space_file.txt", 'w+') as file:
                            file.write(space_file)
                        run_command(appui, SMILES_input, space_file, similarity, min_sim_threshold, num_of_results, query_ID)
                    else:
                        appui.show("Error: " + formatted)
                except Exception as e:
                    print(e)

            progDlg.setProgress(99)
            progDlg.hide()


    except Exception as e:
        log.log(repr(e))

def email_compound(appui):
    """
        This tool creates an email to send to Enamine requesting a quote for the listed compounds.
        Any compounds selected in the data set will be included in the email. The email is created
        in the default email client.


        :param appui:               The stardrop appui object
        :return: None
    """
    try:
        cancelled = False
        # get currently selected molecule
        SMILES_input = appui.getCurrentMolecules()
        if len(SMILES_input) > 350:
            cancelled = True
            appui.show('Please select fewer compounds')


        # if no molecule is selected, tell user to select one and re-run
        if not SMILES_input:
            cancelled = True
            appui.show(
                "Please select at least one molecule before creating the quote request")

        if not cancelled:
            # Import the current data set to get the structures
            dataset = appui.importDS("")
            ds = app.Dataset()
            ds.parse(dataset)

            # Find the structure column
            structureColumn = -1
            result_name_col = -1
            for colIndex in xrange(ds.columnSz()):
                type = ds.columnType(colIndex)
                if ds.columnName(colIndex) == 'REAL space name':
                    result_name_col = colIndex
                if structureColumn == -1 and type == app.mol:
                    structureColumn = colIndex

            if result_name_col == -1:
                appui.show("No result name to send to Enamine, only SMILES string will be included")
                list_of_comps = ""
                for item in SMILES_input:
                    for rowIndex in xrange(ds.rowSz()):
                        SMILES = ds.stringval(structureColumn, rowIndex)
                        if item == SMILES:
                            list_of_comps += 'SMILES: ' + SMILES + '%0D%0D'

            else:
                list_of_comps = ""
                for item in SMILES_input:
                    for rowIndex in xrange(ds.rowSz()):
                        SMILES = ds.stringval(structureColumn, rowIndex)
                        if result_name_col == -1:
                            result_name = ""
                        else:
                            result_name = ds.stringval(result_name_col, rowIndex)
                        if item == SMILES:
                            if "__" in result_name:
                                None
                            else:
                                result_name = result_name.split(">")[1]
                                result_name = result_name.split("<")[0]
                            list_of_comps += 'SMILES: ' + item + '%0DResult name: ' + result_name + '%0D%0D'

            # Create the email containing the list of compounds. If there is no default email client,
            # show and error
            if len(SMILES_input) == 1:
                try:
                    webbrowser.open(
                        'mailto:libraries@enamine.net?subject=REAL space compound request for quote&body=Hello,%0D%0DPlease provide me with a quote for <amount in mg or g> of the following compound:%0D%0D' + list_of_comps + '%0DKind regards,%0D<your name>%0D%0D',
                        new=1)
                except Exception:
                    appui.show('Could not find default email client')
            else:
                try:
                    webbrowser.open('mailto:libraries@enamine.net?subject=REAL space compound request for quote&body=Hello,%0D%0DPlease provide me with a quote for <amount in mg or g> of each of the following compounds:%0D%0D' + list_of_comps + '%0DKind regards,%0D<your name>%0D%0D', new=1)
                except Exception:
                    appui.show(
                        'Could not find default email client')


    except Exception as e:
        log.log(repr(e))


def get_stardrop_definitions():
    definitions = [
        {
            'script_name': 'Enamine REAL Space/Search using selected compound...',
            'callback': search_using_FTrees
        },
        {
            'script_name': 'Enamine REAL Space/Request quote for selected compounds',
            'callback': email_compound
        }
    ]

    return definitions